#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "hdf.h"
#include "mfhdf.h"
#include "vg.h"

#include "hdiff.h"
#include "dumplib.h"
#include "vardata.h"

void
gattr_diff(sdid1, sdid2)
int sdid1, sdid2;
{
  int32 nvars1, nvars2;		/* number of variables */
  int32 ngatts1, ngatts2;	/* number of global attributes */
  struct ncatt att1, att2;	/* attribute */
  int ia, ib;			/* attribute number */
  int iv1, iv2;			/* variable number */
  int iret1, iret2;

  printf ("\n*****     Global attributes comparison:     *****\n");

  iret1 =  SDfileinfo(sdid1, &nvars1, &ngatts1);

  /* get global attributes */

  for (ia = 0; ia < ngatts1; ia++) 
  {
    iret1 = SDattrinfo(sdid1, ia, att1.name, &att1.type, &att1.len);
    ib = SDfindattr(sdid2, att1.name);
    if (ib == -1)     /* attribute doesn't exist in file2 */
    {
      printf("\n---------------------------\n");
      printf("< %s\n", att1.name);
      printf("> '%s' does not exist in file2\n", att1.name);
      continue;
    }
    iret2 = SDattrinfo(sdid2, ib, att2.name, &att2.type, &att2.len);
    att1.val = (void *) malloc((unsigned) (att1.len*DFKNTsize(att1.type)));
    if (!att1.val) 
    {
      error("Out of memory!");
      SDend(sdid1);
      SDend(sdid2);
      exit;
    }
    att2.val = (void *) malloc((unsigned) (att2.len*DFKNTsize(att2.type)));
    if (!att2.val) 
    {
      error("Out of memory!");
      SDend(sdid1);
      SDend(sdid2);
      exit;
    }
    iret1 = SDreadattr(sdid1, ia, att1.val);
    iret2 = SDreadattr(sdid2, ib, att2.val);

    iret2 = 0;
    if (att1.type != att2.type || att1.len != att2.len) 
      iret2 = 1;
    if (iret2 == 0)       /* compare the data */
      iret2 = memcmp(att1.val, att2.val, att1.len*DFKNTsize(att1.type));

    if (iret2 != 0)
    {
      printf("\n---------------------------\n");
      printf("Attr Name: %s\n", att1.name);
      printf("< ");
      pr_att_vals(att1.type, att1.len, att1.val);
      printf("\n> ");
      pr_att_vals(att2.type, att2.len, att2.val);
      printf("\n");
    }

    free ((char *) att1.val);
    free ((char *) att2.val);
  }

  /* check any global attributes in file2 but not in file1 */

  iret2 =  SDfileinfo(sdid2, &nvars2, &ngatts2);

  for (ib = 0; ib < ngatts2; ib++) 
  {
    iret2 = SDattrinfo(sdid2, ib, att2.name, &att2.type, &att2.len);
    ia = SDfindattr(sdid1, att2.name);
    if (ia == -1)     /* attribute doesn't exist in file2 */
    {
      printf("\n---------------------------\n");
      printf("< '%s' does not exist in file1\n", att2.name);
      printf("> %s\n", att2.name);
    }
  }
 
}


void
sdattr_diff(sdid1, sdid2)
int sdid1, sdid2;
{
  int32 nvars1, nvars2;		/* number of variables */
  int32 ngatts1, ngatts2;	/* number of global attributes */
  int varid1, varid2;		/* variable id */
  struct ncvar var1, var2;	/* variable */
  struct ncatt att1, att2;	/* attribute */
  int id;			/* dimension number per variable */
  int ia, ib;			/* attribute number */
  int iv1, iv2;			/* variable number */
  int iret1, iret2, iflag, istop;

  printf ("\n*****     SD attributes comparison:     *****\n");

  iret1 =  SDfileinfo(sdid1, &nvars1, &ngatts1);
  iret2 =  SDfileinfo(sdid2, &nvars2, &ngatts2);

  /* check any sd in file2 but not in file1 */

  for (iv2 = 0; iv2 < nvars2; iv2++) 
  {
    varid2 = SDselect(sdid2, iv2);
    iret2 = SDgetinfo(varid2, var2.name, &var2.ndims, var2.dims, &var2.type,
                      &var2.natts);

    iv1 = SDnametoindex(sdid1, var2.name);
    if (iv1 == -1)     /* sd doesn't exist in file2 */
    {
      SDendaccess(varid2);
      printf("\n---------------------------\n");
      printf("< '%s' does not exist in file1\n", var2.name);
      printf("> %s\n", var2.name);
      continue;
    }
  }

  /* complete comparison of file1 to file2 */

  for (iv1 = 0; iv1 < nvars1; iv1++) 
  {
    varid1 = SDselect(sdid1, iv1);
    iret1 = SDgetinfo(varid1, var1.name, &var1.ndims, var1.dims, &var1.type,
                      &var1.natts);

    iv2 = SDnametoindex(sdid2, var1.name);
    if (iv2 == -1)     /* sd doesn't exist in file2 */
    {
      SDendaccess(varid1);
      printf("\n---------------------------\n");
      printf("< %s\n", var1.name);
      printf("> '%s' does not exist in file2\n", var1.name);
      continue;
    }

    varid2 = SDselect(sdid2, iv2);
    iret2 = SDgetinfo(varid2, var2.name, &var2.ndims, var2.dims, &var2.type,
                      &var2.natts);

    iflag = 0;
    if (var1.ndims != var2.ndims || var1.type != var2.type || 
        var1.natts != var2.natts)
      iflag = 1;

    if (iflag == 0)
    {
      for (id = 0; id < var1.ndims; id++)
        if (var1.dims[id] != var2.dims[id])
          iflag = 1;
    }

    if (iflag == 0)    /* check attributes of the sd */
    {
      istop = 0;
      for (ia = 0; ia < var1.natts, istop = 0; ia++) 
      {
        iret1 = SDattrinfo(varid1, ia, att1.name, &att1.type, &att1.len);
        ib = SDfindattr(varid2, att1.name);
        if (ib == -1)
        {
          istop = 1;
          iflag = 1;
          continue;
        }
        iret2 = SDattrinfo(varid2, ib, att2.name, &att2.type, &att2.len);
        if (att1.type != att2.type || att1.len != att2.len)
        {
          istop = 1;
          iflag = 1;
          continue;
        }

        att1.val = (void *) malloc((unsigned)att1.len*DFKNTsize(att1.type));
        if (!att1.val) 
        {
          error("Out of memory!");
          SDend(sdid1);
          SDend(sdid2);
          exit;
        }
        att2.val = (void *) malloc((unsigned) (att2.len*DFKNTsize(att2.type)));
        if (!att2.val)
        {
          error("Out of memory!");
          SDend(sdid1);
          SDend(sdid2);
          exit;
        }
        iret1 = SDreadattr(varid1, ia, att1.val);
        iret2 = SDreadattr(varid2, ib, att2.val);

        iret1 = memcmp(att1.val, att2.val, att1.len*DFKNTsize(att1.type));
        if (iret1 != 0)
          iflag = 1;
 
        free ((char *) att1.val);
        free ((char *) att2.val);
      }
    }

    if (iflag == 0)    /* check if any attributes in var2 but not in var1 */
    {
      istop = 0;
      for (ib = 0; ib < var2.natts, istop = 0; ib++) 
      {
        iret2 = SDattrinfo(varid2, ib, att2.name, &att2.type, &att2.len);
        ia = SDfindattr(varid1, att2.name);
        if (ia == -1)
        {
          istop = 1;
          iflag = 1;
          continue;
        }
      }
    }

    if (iflag == 1)    /* print out the difference */
    {
      printf("\n---------------------------\n");
      printf("SD Name: %s\n", var1.name);
      printf("<<<<\n");
    
      if (var1.ndims > 0)
        Printf ("(");

      for (id = 0; id < var1.ndims; id++) 
        Printf ("%d%s", var1.dims[id], id < var1.ndims-1 ? ", " : ")");

      Printf (" ;\n");
      /* get variable attributes */
      for (ia = 0; ia < var1.natts; ia++) 
      {
        iret1 = SDattrinfo(varid1, ia, att1.name, &att1.type, &att1.len);

        Printf ("\t%s:%s = ", var1.name, att1.name);

        att1.val = (void *) malloc((unsigned)att1.len*DFKNTsize(att1.type));
        if (!att1.val) 
        {
          error("Out of memory!");
          SDend(sdid1);
          SDend(sdid2);
          exit;
        }
        iret1 = SDreadattr(varid1, ia, att1.val);
        pr_att_vals(att1.type, att1.len, att1.val);
        Printf (" ;\n");
        free ((char *) att1.val);
      }
      printf(">>>>\n");
    
      if (var2.ndims > 0)
        Printf ("(");

      for (id = 0; id < var2.ndims; id++) 
        Printf ("%d%s", var2.dims[id], id < var2.ndims-1 ? ", " : ")");

      Printf (" ;\n");
      /* get variable attributes */
      for (ia = 0; ia < var2.natts; ia++) 
      {
        iret1 = SDattrinfo(varid2, ia, att2.name, &att2.type, &att2.len);

        Printf ("\t%s:%s = ", var2.name, att2.name);

        att2.val = (void *) malloc((unsigned)att2.len*DFKNTsize(att2.type));
        if (!att2.val) 
        {
          error("Out of memory!");
          SDend(sdid1);
          SDend(sdid2);
          exit;
        }
        iret2 = SDreadattr(varid2, ia, att2.val);
        pr_att_vals(att2.type, att2.len, att2.val);
        Printf (" ;\n");
        free ((char *) att2.val);
      }
    }
    SDendaccess(varid1);
    SDendaccess(varid2);
  }
}



void
sddata_diff(sdid1, sdid2, specp)
int sdid1, sdid2;
struct fspec specp;
{
  int32 nvars1, nvars2;		/* number of variables */
  int32 ngatts1, ngatts2;	/* number of global attributes */
  int varid1, varid2;		/* variable id */
  struct ncvar var1, var2;	/* variable */
  struct ncatt att1, att2;	/* attribute */
  int id;			/* dimension number per variable */
  int ia, ib;			/* attribute number */
  int iv1, iv2;			/* variable number */
  int i, k, iret1, iret2, iflag, tot_cnt;
  vnode* vlist = newvlist();  /* list for vars specified with -v option */
  int32 start[MAX_NC_DIMS];
  void *buf1, *buf2, array_diff();

  /*
   * If any vars were specified with -v option, get list of associated
   * variable ids
   */
  for (k=0; k < specp.nlvars; k++) 
  {
    varid1 = SDnametoindex(sdid1, specp.lvars[k]);
    varadd(vlist, varid1);
  }

  printf ("\n*****     SD Data comparison:     *****\n");

  iret1 =  SDfileinfo(sdid1, &nvars1, &ngatts1);

  for (k = 0; k < nvars1; k++)
  {
    varid1 = SDselect(sdid1, k);

    /* if var list specified, test for membership */
    if (specp.nlvars > 0 && ! varmember(vlist, k))
      continue;

    iret1 = SDgetinfo(varid1, var1.name, &var1.ndims, var1.dims, &var1.type,
                    &var1.natts);

    iv2 = SDnametoindex(sdid2, var1.name);

    if (iv2 == -1)     /* sd doesn't exist in file2 */
    {
      SDendaccess(varid1);
      continue;
    }

    varid2 = SDselect(sdid2, iv2);
    iret2 = SDgetinfo(varid2, var2.name, &var2.ndims, var2.dims, &var2.type,
                      &var2.natts);

    if (var1.ndims != var2.ndims || var1.type != var2.type) 
    {
      SDendaccess(varid1);
      continue;
    }

    printf("\n---------------------------\n");
    printf("SD Name: %s .... Comparing\n", var1.name);

    tot_cnt = 1;

    for (iv1 = 0; iv1 < var1.ndims; iv1++) 
    {
      start[iv1] = 0;
      tot_cnt *= var1.dims[iv1];
    }
      
    buf1 = (void *) malloc((unsigned) (tot_cnt*DFKNTsize(var1.type)));
    buf2 = (void *) malloc((unsigned) (tot_cnt*DFKNTsize(var1.type)));

    iret1 = SDreaddata(varid1, start, (int32 *)NULL, var1.dims, buf1);
    iret2 = SDreaddata(varid2, start, (int32 *)NULL, var1.dims, buf2);
 
    array_diff(buf1, buf2, tot_cnt, var1.type, specp.max_err_cnt);
    free((char *) buf1);
    free((char *) buf2);

    SDendaccess(varid1);
    SDendaccess(varid2);
  }
}




void
vdata_diff(fid1, fid2, specp)
int fid1, fid2;
struct fspec specp;
{
  int32 vsid1, vsid2;
  int32 vs1, vs2;
  int32 i;

  int32 nv1, interlace1, vsize1;
  int32 vsotag1, vsoref1;
  char fields1[50], vsname1[50];
  char  vgclass1[50],vsclass1[50], *name1;
  int32 nv2, interlace2, vsize2;
  int32 vsotag2, vsoref2;
  char fields2[50], vsname2[50];
  char  vgclass2[50],vsclass2[50], *name2;
  void *buf1, *buf2;

  Vstart(fid1);
  Vstart(fid2);

  printf ("\n*****     Vdata comparison:     *****\n");
/*
  check any vdata in file1, but not in file2
*/   
/*
  vsid1 = -1;
  while ( (vsid1 = VSgetid(fid1, vsid1)) != -1)
  {
    vs1 = VSattach(fid1, vsid1, "r");
    if (vs1 == FAIL)
    {
      printf("Cannot open file1 vs id = %d\n", vsid1);
      continue;
    } 
    VSinquire(vs1, &nv1, &interlace1, fields1, &vsize1, vsname1);

    vsoref2 = VSfind(fid2, vsname1);
    if (vsoref2 <= 0)
    { 
      printf("\n---------------------------\n");
      printf("< Vdata '%s'\n", vsname1);
      printf("> Vdata '%s' does not exist in file2\n", vsname1);
      VSdetach(vs1);
      continue;
    }
    VSdetach(vs1);
  }
*/
/*
  check any vdata in file2, but not in file1
*/   
/*
  vsid2 = -1;
  while ( (vsid2 = VSgetid(fid2, vsid2)) != -1)
  {
    vs2 = VSattach(fid2, vsid2, "r");
    if (vs2 == FAIL)
    {
      printf("Cannot open file2 vs id = %d\n", vsid2);
      continue;
    } 
    VSinquire(vs2, &nv2, &interlace2, fields2, &vsize2, vsname2);

    vsoref1 = VSfind(fid1, vsname2);
    if (vsoref1 <= 0)
    { 
      printf("\n---------------------------\n");
      printf("< Vdata '%s' does not exist in file1\n", vsname2);
      printf("> Vdata '%s'\n", vsname2);
      VSdetach(vs2);
      continue;
    }
    VSdetach(vs2);
  }
*/
/*
  doing comparison based on file1
*/
  vsid1 = -1;
  while ( (vsid1 = VSgetid(fid1, vsid1)) != -1)
  {
    vs1 = VSattach(fid1, vsid1, "r");
    if (vs1 == FAIL)
      continue;

    VSinquire(vs1, &nv1, &interlace1, fields1, &vsize1, vsname1);

    vsoref2 = VSfind(fid2, vsname1);
    if (vsoref2 <= 0)   /* skip, if vsname1 is not in file2 */
    {
      VSdetach(vs1);
      continue;
    }
    vs2 = VSattach(fid2, vsoref2, "r");

    VSinquire(vs2, &nv2, &interlace2, fields2, &vsize2, vsname2);

    vsotag1 = VSQuerytag(vs1);
    vsoref1 = VSQueryref(vs1);
    VSgetclass(vs1,vsclass1);

    vsotag2 = VSQuerytag(vs2);
    vsoref2 = VSQueryref(vs2);
    VSgetclass(vs2,vsclass2);

    if (vsotag1 != vsotag2 || nv1 != nv2 || interlace1 != interlace2 ||
        strcmp(fields1, fields2) != 0 || strcmp(vsclass1, vsclass2) != 0 ||
        strcmp(vsclass1, "Attr0.0") != 0 && vsize1 != vsize2)
    {
      printf("\n---------------------------\n");
      printf("Vdata Name: %s\n", vsname1);
      printf("< <%d> nrec=%d interlace=%d fld=[%s] vsize=%d class={%s})\n",
             vsotag1, nv1, interlace1, fields1, vsize1, vsclass1);
      printf("> <%d> nrec=%d interlace=%d fld=[%s] vsize=%d class={%s})\n",
             vsotag2, nv2, interlace2, fields2, vsize2, vsclass2);
      VSdetach(vs1);
      VSdetach(vs2);
      continue;     /* Don't compare the data */ 
    }
  
    /* compare the data */

    buf1 = (void *) malloc((unsigned) (nv1 * vsize1));
    buf2 = (void *) malloc((unsigned) (nv2 * vsize2));
    if (!buf1 || !buf2) 
    {
      error("Out of memory!");
      VSdetach(vs1); Hclose(fid1);
      VSdetach(vs2); Hclose(fid2);
      exit;
    }

    VSsetfields(vs1, fields1);
    VSread(vs1, buf1, nv1, interlace1);
    VSsetfields(vs2, fields2);
    VSread(vs2, buf2, nv2, interlace2);

    for (i=0; i<nv1; i++)
    {
      if (vsize1 == vsize2)
      {
        
      }
      else
      {
      }
    }

    VSdetach(vs1);
    VSdetach(vs2);
    free((char *) buf1);
    free((char *) buf2);

  }

  Vend(fid1);
  Vend(fid2);
}




void
array_diff(buf1, buf2, tot_cnt, type, max_err_cnt)
void *buf1, *buf2;
int32 type, tot_cnt, max_err_cnt;
{
  int32 i, err_cnt;
  char *i1ptr1, *i1ptr2;
  short *i2ptr1, *i2ptr2;
  long  *i4ptr1, *i4ptr2;
  float *fptr1, *fptr2;
  double *dptr1, *dptr2;

  err_cnt = 0;

  switch(type)
  {
    case DFNT_INT8:
      i1ptr1 = (char *) buf1;
      i1ptr2 = (char *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (*i1ptr1 != *i1ptr2)
        {
          printf("Index: %ld,   File1: %d,   File2: %d\n",
                 i, *i1ptr1, *i1ptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        i1ptr1++;  i1ptr2++;
      }

      break;

    case DFNT_INT16:
      i2ptr1 = (short *) buf1;
      i2ptr2 = (short *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (*i2ptr1 != *i2ptr2)
        {
          printf("Index: %ld,   File1: %d,   File2: %d\n",
                 i, *i2ptr1, *i2ptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        i2ptr1++;  i2ptr2++;
      }

      break;

    case DFNT_INT32:
      i4ptr1 = (long *) buf1;
      i4ptr2 = (long *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (*i4ptr1 != *i4ptr2)
        {
          printf("Index: %ld,   File1: %ld,   File2: %ld\n",
                 i, *i4ptr1, *i4ptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        i4ptr1++;  i4ptr2++;
      }

      break;

    case DFNT_FLOAT:
      fptr1 = (float *) buf1;
      fptr2 = (float *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (*fptr1 != *fptr2)
        {
          printf("Index: %ld,   File1: %#.8g,   File2: %#.8g\n",
                 i, *fptr1, *fptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        fptr1++;  fptr2++;
      }

      break;

    case DFNT_DOUBLE:
      dptr1 = (double *) buf1;
      dptr2 = (double *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (*dptr1 != *dptr2)
        {
          printf("Index: %ld,   File1: %#.16g,   File2: %#.16g\n",
                 i, *dptr1, *dptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }
        dptr1++;  dptr2++;
      }

      break;

    default:
      error(" bad type - %d\n", type);
  }

}
