#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>

#include "hdf.h"
#include "mfhdf.h"
#include "vg.h"

#include "hdiff.h"
#include "dumplib.h"
#include "vardata.h"

void
sddata_diff(sdid1, sdid2, specp)
int sdid1, sdid2;
struct fspec specp;
{
  int32 nvars1, nvars2;		/* number of variables */
  int32 ngatts1, ngatts2;	/* number of global attributes */
  int varid1, varid2;		/* variable id */
  struct ncvar var1, var2;	/* variable */
  struct ncatt att1, att2;	/* attribute */
  int id;			/* dimension number per variable */
  int ia, ib;			/* attribute number */
  int iv1, iv2;			/* variable number */
  int i, k, iret1, iret2, iflag, tot_cnt, tot_cnt1, tot_cnt2;
  vnode* vlist = newvlist();  /* list for vars specified with -v option */
  int32 start[MAX_NC_DIMS];
  void *buf1, *buf2, array_diff();

  /*
   * If any vars were specified with -v option, get list of associated
   * variable ids
   */
  for (k=0; k < specp.nlvars; k++) 
  {
    varid1 = SDnametoindex(sdid1, specp.lvars[k]);
    varadd(vlist, varid1);
  }

  printf ("\n*****     SD Data comparison:     *****\n");

  iret1 =  SDfileinfo(sdid1, &nvars1, &ngatts1);

  for (k = 0; k < nvars1; k++)
  {
    varid1 = SDselect(sdid1, k);

    /* if var list specified, test for membership */
    if (specp.nlvars > 0 && ! varmember(vlist, k))
      continue;

    iret1 = SDgetinfo(varid1, var1.name, &var1.ndims, var1.dims, &var1.type,
                    &var1.natts);

    iv2 = SDnametoindex(sdid2, var1.name);

    if (iv2 == -1)     /* sd doesn't exist in file2 */
    {
      SDendaccess(varid1);
      continue;
    }

    varid2 = SDselect(sdid2, iv2);
    iret2 = SDgetinfo(varid2, var2.name, &var2.ndims, var2.dims, &var2.type,
                      &var2.natts);

    if (var1.ndims != var2.ndims || var1.type != var2.type) 
    {
      SDendaccess(varid1);
      continue;
    }

    printf("\n---------------------------\n");
    printf("SD Name: %s .... Comparing\n", var1.name);

    tot_cnt1 = 1;
    for (iv1 = 0; iv1 < var1.ndims; iv1++) 
    {
      start[iv1] = 0;
      tot_cnt1 *= var1.dims[iv1];
    }
      
    tot_cnt2 = 1;
    for (iv1 = 0; iv1 < var2.ndims; iv1++) 
    {
      start[iv1] = 0;
      tot_cnt2 *= var2.dims[iv1];
    }
      
    buf1 = (void *) malloc((unsigned) (tot_cnt1*DFKNTsize(var1.type)));
    buf2 = (void *) malloc((unsigned) (tot_cnt2*DFKNTsize(var2.type)));

    iret1 = SDreaddata(varid1, start, (int32 *)NULL, var1.dims, buf1);
    iret2 = SDreaddata(varid2, start, (int32 *)NULL, var2.dims, buf2);
 
/*  in case there is a difference in the elements, compare up to the
    limit of the less one */
 
    if (tot_cnt1 > tot_cnt2)
      tot_cnt = tot_cnt2;
    else
      tot_cnt = tot_cnt1; 

    array_diff(buf1, buf2, tot_cnt, var1.type, specp.err_limit, 
               specp.max_err_cnt);
    free((char *) buf1);
    free((char *) buf2);

    SDendaccess(varid1);
    SDendaccess(varid2);
  }
}



void
array_diff(buf1, buf2, tot_cnt, type, err_limit, max_err_cnt)
void *buf1, *buf2;
int32 type, tot_cnt, max_err_cnt;
float err_limit;
{
  int32 i, err_cnt;
  char *i1ptr1, *i1ptr2;
  short *i2ptr1, *i2ptr2;
  long  *i4ptr1, *i4ptr2;
  float *fptr1, *fptr2;
  double *dptr1, *dptr2;

  err_cnt = 0;

  switch(type)
  {
    case DFNT_INT8:
    case DFNT_UINT8:
      i1ptr1 = (char *) buf1;
      i1ptr2 = (char *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (abs(*i1ptr1 - *i1ptr2) > (int) err_limit)
        {
          printf("Index: %ld,   File1: %d,   File2: %d\n",
                 i, *i1ptr1, *i1ptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        i1ptr1++;  i1ptr2++;
      }

      break;

    case DFNT_INT16:
    case DFNT_UINT16:
      i2ptr1 = (short *) buf1;
      i2ptr2 = (short *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (abs(*i2ptr1 - *i2ptr2) > (int) err_limit)
        {
          printf("Index: %ld,   File1: %d,   File2: %d\n",
                 i, *i2ptr1, *i2ptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        i2ptr1++;  i2ptr2++;
      }

      break;

    case DFNT_INT32:
    case DFNT_UINT32:
      i4ptr1 = (long *) buf1;
      i4ptr2 = (long *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (labs(*i4ptr1 - *i4ptr2) > (long) err_limit)
        {
          printf("Index: %ld,   File1: %ld,   File2: %ld\n",
                 i, *i4ptr1, *i4ptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        i4ptr1++;  i4ptr2++;
      }

      break;

    case DFNT_FLOAT:
      fptr1 = (float *) buf1;
      fptr2 = (float *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (fabs(*fptr1 - *fptr2) > err_limit)
        {
          printf("Index: %ld,   File1: %#.8g,   File2: %#.8g\n",
                 i, *fptr1, *fptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }                                               
        fptr1++;  fptr2++;
      }

      break;

    case DFNT_DOUBLE:
      dptr1 = (double *) buf1;
      dptr2 = (double *) buf2;
      for (i=0; i<tot_cnt; i++)
      {
        if (fabs(*dptr1 - *dptr2) > (double) err_limit)
        {
          printf("Index: %ld,   File1: %#.16g,   File2: %#.16g\n",
                 i, *dptr1, *dptr2);
          if (max_err_cnt > 0)
          {
            err_cnt++;
            if (err_cnt >= max_err_cnt)
              break;
          }
        }
        dptr1++;  dptr2++;
      }

      break;

    default:
      error(" bad type - %d\n", type);
  }

}
